
{{define "title"}}
  Go Application | {{index . "name"}}
{{end}}

{{define "body"}}
        
        <h1 class="hello" style='color:red;'>Sample Application !!!!!</h1>
        <header >
        <link rel="stylesheet" type="text/css" href="/static/css/main.css">
        <img src="/static/image/goimg.png" alt ="Gopher">
        <b style='color:red; !important;'>Cavisson System </b>
        </header>
        <nav>
        <ul>
            <li style='color:red;'><a href="/"; style='color:lightyellow;'> Home </a></li>
            <li style='color:red;'><a href="/page"; style='color:lightyellow;'>Content </a></li>
            <li style='color:red;'><a href="/about"; style='color:lightyellow;'>About</a></li>
        </ul>
        </nav> 
        <hr>
    <p style='color:red;'> <b> Normalcall <a href="/hello" ; style='color:lightyellow;'>Page render</a> </p>
    <p style='color:red;'> <b> Mysql     <a href="/mysql"; style='color:lightyellow;'>Mysql Page </a></p>
    <p style='color:red;'> <b> Cats      <a href="/cats?name=Abhi&type=data"; style='color:lightyellow;'>cat Page </a></p>
    <p style='color:red;'> <b> Json      <a href="/json/abhi/karan/abhay"; style='color:lightyellow;'>json Page </a></p>
    <p style='color:red;'> <b>  Mongodb   <a href="/mongodb"; style='color:lightyellow;'>Mongodb Page </a></p>
    <p style='color:red;'> <b> Cassandra <a href="/cassandra"; style='color:lightyellow;'>cassandra Page </a></p>
    <p style='color:red;'> <b> Postgres  <a href="/postgres"; style='color:lightyellow;'>postgres Page </a></p>
    <p style='color:red;'> <b> Postpg    <a href="/postpg"; style='color:lightyellow;'>postgpg Page </a></p>
    <p style='color:red;'> <b> gorilla <a href="/mux/"; style='color:lightyellow;'>Gorilla Page </a></p>
    <p style='color:red;'> <b> http callout   <a href="/httpcall"; style='color:lightyellow;'>http_callout</a></p>
{{end}}
