{{define "title"}}
  Go Application | {{index . "name"}}
{{end}}

{{define "body"}}
        <header>
        <link rel="stylesheet" type="text/css" href="/static/css/main.css">
        <img src="/static/image/goimg.png" alt ="Gopher">
        <b style='color:red; !important;'>Cavisson System </b>
        </header>
        <nav>
        <ul>
            <li><a href="/"> Home </a></li>
            <li><a href="/page">Content </a></li>
            <li><a href="/about">About</a></li>
        </ul>
        </nav> 
        <hr>

        <h1>Http Callout Successful</h1>

{{end}}