{{define "title"}}
  Go Application | {{index . "name"}}
{{end}}

{{define "body"}}
        <header>
        <link rel="stylesheet" type="text/css" href="/static/css/main.css">
        <img src="/static/image/goimg.png" alt ="Gopher">
        <b style='color:red !important;'>Cavisson System </b>
        </header>
        <nav>
        <ul>
            <li style='color:red;'><a href="/"; style='color:lightyellow;'> Home </a></li>
            <li style='color:red;'><a href="/page"; style='color:lightyellow;'>Content </a></li>
            <li style='color:red;'><a href="/about"; style='color:lightyellow;'>About</a></li>
        </ul>
        </nav> 
        <hr>
        <h1>Json Successful</h1>
{{end}}
