package main

import (
	"fmt"
	"io/ioutil"
	"net/http"
	"os"
)

func main() {
	if len(os.Args) < 2 {
		//fmt.Println("Usage: go run main.go <URL>")
		return
	}

	url := os.Args[1]

	resp, err := http.Get(url)
	if err != nil {
		//fmt.Printf("Error downloading file: %v", err)
		return
	}
	defer resp.Body.Close()

	if resp.StatusCode != http.StatusOK {
		//fmt.Printf("Error downloading file: %s", resp.Status)
		return
	}

	content, err := ioutil.ReadAll(resp.Body)
	if err != nil {
		//fmt.Printf("Error reading response body: %v", err)
				return
	}

	fmt.Println(string(content))
}

