#!/bin/bash

if [[ $# -ne 2 ]]; then
  echo "Usage: $0 <filename> <key>"
  exit 1
fi

filename=$1
key=$2
CUR_DIR=$( cd "$(dirname "$0")" ; pwd -P )
#DISTRO=$($CUR_DIR/nsi_get_linux_release_ex -d)
jqpath=$CUR_DIR/jq_64
value=$($jqpath -r .$key $filename)
echo "$value"

