#!/bin/bash
CUR_DIR=$( cd "$(dirname "$0")" ; pwd -P )
DISTRO=$($CUR_DIR/nsi_get_linux_release_ex -d)
RELEASE=$($CUR_DIR/nsi_get_linux_release_ex -r)
#u20path=$CUR_DIR/../../../../ubuntu-pkg/Ubuntu_2004/pkgs/packages
if [[ $DISTRO == "Ubuntu" &&  $RELEASE -eq 1604 ]];then
	echo "installing curl on ubuntu16"
	#u16path=$CUR_DIR/../../../../ubuntu-pkg/Ubuntu_1604/pkgs/packages
elif [[ $DISTRO == "Ubuntu" &&  $RELEASE -eq 2004 ]];then
	echo "installing curl on ubuntu20"
elif [ "$DISTRO" == "Redhat" ]; then
	echo "installing curl on redhat"
fi
