/*
 * A toy library to demonstrate the codeCoverage tool
 */

#include "libwmalloc.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

long fastMalloc(long x)
{
    fprintf(stderr,"\nin fastMalloc %ld\n",x);
    sleep(1);
    fprintf(stderr,"return from fastMalloc\n");
    return x + 5;
}

void testMalloc() {
fprintf(stderr,"\nin testMalloc\n");
    sleep(1);
    fprintf(stderr,"return from testMalloc\n");
    }

