#include "libusejson.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "parson.h"

void TestJson() {
	JSON_Value *root_value = json_value_init_object();
  if(!root_value)
	  return ;
  JSON_Object *root_object = json_value_get_object(root_value);
  if(!root_object)
	 return ;
	fprintf(stderr, "json success\n");
}